package no.twofingers.iControl;

import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.io.IOException;
import java.io.FileFilter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;

/**
 * This is the main iControl class
 *
 * @author <a href="mailto:aslak.hellesoy@bekk.no">Aslak Hellesy</a>
 */
public class IControl {

//	private Map allSettings = new HashMap();
	private Map allSettingsByQualifielName = new HashMap();
	private DBCProject dbcProject;
	private File controlFile;
	private File targetsFile;
	private Properties properties;
	private File propsFile;
	private String classpath;

	/**
	 * Analyzes the shebang
	 */
	public void analyze( File propsFile ) {
		this.propsFile = propsFile;
		properties = new Properties();

		classpath = null;
		File classRoot = null;
		String srcRoot = null;
		try {
			properties.load( new FileInputStream( propsFile ) );
			try {
				classpath = properties.getProperty( "classpath" );
			} catch( NullPointerException e ) {
				System.err.println( "Couldn't read property classpath from " + propsFile.getAbsoluteFile() );
				System.exit( 1 );
			}
			try {
				targetsFile = new File( properties.getProperty( "targetsFile" ) );
			} catch( NullPointerException e ) {
				System.err.println( "Couldn't read property targetsFile from " + propsFile.getAbsoluteFile() );
				System.exit( 1 );
			}
			try {
				controlFile = new File( properties.getProperty( "controlFile" ) );
			} catch( NullPointerException e ) {
				System.err.println( "Couldn't read property controlFile from " + propsFile.getAbsoluteFile() );
				System.exit( 1 );
			}
			try {
				classRoot = new File( properties.getProperty( "classRoot" ) );
			} catch( NullPointerException e ) {
				System.err.println( "Couldn't read property classRoot from " + propsFile.getAbsoluteFile() );
				System.exit( 1 );
			}
			try {
				srcRoot = properties.getProperty( "sourceRoot" );
			} catch( NullPointerException e ) {
				System.err.println( "Couldn't read property sourceRoot from " + propsFile.getAbsoluteFile() );
				System.exit( 1 );
			}
		} catch( IOException e ) {
			System.err.println( "Couldn't read properties file from " + propsFile.getAbsoluteFile() );
			System.err.println( "Consider using updateicontrol=\"true\" from your iContract task." );
			System.err.println( "This will create a default file for you." );
			System.exit( 1 );
		}

		DBCClass.setClasspath( classpath );
		DBCClass.setSourceRoot( srcRoot );
		dbcProject = new DBCProject( classRoot );

		// find out how many there are in total

		// populate the allSettings Map, for quicker lookup
		Iterator packages = dbcProject.getChildren().iterator();
		while( packages.hasNext() ) {
			DBCPackage dbcPackage = (DBCPackage) packages.next();
//			addSetting( dbcPackage );
			addSettingByQualifieName( dbcPackage );

			Iterator classes = dbcPackage.getChildren().iterator();
			while( classes.hasNext() ) {
				DBCClass dbcClass = (DBCClass) classes.next();
//				addSetting( dbcClass );
				addSettingByQualifieName( dbcClass );

				Iterator methods = dbcClass.getChildren().iterator();
				while( methods.hasNext() ) {
					DBCMethod dbcMethod = (DBCMethod) methods.next();
//					addSetting( dbcMethod );
					addSettingByQualifieName( dbcMethod );
				}
			}
		}

		load();
	}

	public List getErrorClasses() {
		return Setting.getErrorClasses();
	}

	public String getClasspath() {
		return classpath;
	}

	public Properties getProperties() {
		return properties;
	}

	public final DBCProject getProject() {
		return dbcProject;
	}
/*
	private final void addSetting( Setting setting ) {
		allSettings.put( "" + setting.getId() , setting );
	}
*/
	/*
	private final Setting getSetting( String id ) {
		return (Setting) allSettings.get( id );
	}
	*/

	private final void addSettingByQualifieName( Setting setting ) {
		allSettingsByQualifielName.put( setting.getQualifiedName() , setting );
	}


	private final Setting getSettingByQualifiedName( String qualifiedName ) {
		return (Setting) allSettingsByQualifielName.get( qualifiedName );
	}

	public void load() {
		// read settings file and set settings
		if( controlFile.exists() ) {
			//System.out.print( "Analyzing control file : " );
			try {
				FileReader fr = new FileReader( controlFile );
				BufferedReader br = new BufferedReader( fr );
				String line = null;
				while( ( line = br.readLine() ) != null ) {
					//System.out.print( "." );
					StringTokenizer st = new StringTokenizer( line );
					if( st.hasMoreTokens() ) {
						String qualifiedName = st.nextToken();
						Setting setting = getSettingByQualifiedName( qualifiedName );
						if( st.hasMoreTokens() && setting != null) {
							String settings = st.nextToken();
							if( settings.indexOf( "pre" ) != -1 ) {
								setting.setSetting( Setting.PRE, Setting.ON );
							}
							if( settings.indexOf( "post" ) != -1 ) {
								setting.setSetting( Setting.POST, Setting.ON );
							}
							if( settings.indexOf( "inv" ) != -1 ) {
								setting.setSetting( Setting.INVARIANT, Setting.ON );
							}
						}
					}
				}
			} catch( Exception e ) {
				e.printStackTrace();
				// TODO: log it
			}
			//System.out.println( "Done." );
		} else {
			//Logger.get().info( "No control file to analyze." );
		}
	}

	public void saveAssertionSettings() throws IOException {
		FileWriter fw = new FileWriter( controlFile );
		PrintWriter pw = new PrintWriter( fw );

		FileWriter targetsFileWriter = new FileWriter( targetsFile );
		PrintWriter targetsPrintWriter = new PrintWriter( targetsFileWriter );

		boolean didWrite = false;

		// populate the allSettings Map, for quicker lookup
		Iterator packages = dbcProject.getChildren().iterator();
		while( packages.hasNext() ) {
			DBCPackage dbcPackage = (DBCPackage) packages.next();
			Iterator classes = dbcPackage.getChildren().iterator();
			while( classes.hasNext() ) {
				DBCClass dbcClass = (DBCClass) classes.next();
				// check if the class has any assertions enabled
				boolean classHasPre = !dbcClass.getSetting( Setting.PRE ).equals( Setting.OFF );
				boolean classHasPost = !dbcClass.getSetting( Setting.POST ).equals( Setting.OFF );
				boolean classHasInvariant = !dbcClass.getSetting( Setting.INVARIANT ).equals( Setting.OFF );
				boolean classHasAssertions = classHasPre || classHasPost || classHasInvariant;

				if( classHasAssertions ) {
					// the class has some assertions enabled

					// write class' source file name to targets file
					targetsPrintWriter.println( dbcClass.getSourceFileName() );

					// write class' assertion settings to control file
//					if( classHasInvariant ) {
						// only print classes that have inv enabled (they will never print pre,post)
						pw.println( dbcClass.getControlFileName() );
						didWrite = true;
//					}

					Iterator methods = dbcClass.getChildren().iterator();
					while( methods.hasNext() ) {
						// check if the method has any assertions enabled
						DBCMethod dbcMethod = (DBCMethod) methods.next();

						boolean methodHasPre = dbcMethod.getSetting( Setting.PRE ).equals( Setting.ON );
						boolean methodHasPost = dbcMethod.getSetting( Setting.POST ).equals( Setting.ON );

						if( methodHasPre || methodHasPost ) {
							pw.println( dbcMethod.getControlFileName() );
							didWrite = true;
						}
					}
				}
			}
		}
		pw.flush();
		pw.close();
		targetsPrintWriter.flush();
		targetsPrintWriter.close();
		if( didWrite ) {
			System.out.println( "Saved " + controlFile.getAbsolutePath() );
		} else {
			controlFile.delete();
			System.out.println( "Deleted " + controlFile.getAbsolutePath() );
		}
	}

	public void saveProperties() throws IOException {
		properties.store( new FileOutputStream( propsFile ), "Generated by iControl" );
		System.out.println( "Saved " + propsFile.getAbsolutePath() );
	}

	public void save() throws IOException {
		saveProperties();
		saveAssertionSettings();
	}

}
