package no.twofingers.iControl;

import java.util.Iterator;
import java.util.List;
import java.util.Collection;
import java.util.ArrayList;
import java.io.FileFilter;
import java.io.File;

/**
 * Defines the possible values for assertion settings
 *
 * @author  <a href="mailto:aslak.hellesoy@bekk.no">Aslak Hellesy</a>
 *
 * @invariant getSetting( PRE ) == ON || getSetting( PRE ) == OFF || getSetting( PRE) == MIXED
 * @invariant getSetting( POST ) == ON || getSetting( POST ) == OFF || getSetting( POST) == MIXED
 * @invariant getSetting( INVARIANT ) == ON || getSetting( INVARIANT ) == OFF || getSetting( INVARIANT) == MIXED
 *
 * @invariant forall Setting s1 initerator getChildren().iterator() |
 *	              forall Setting s2 initerator getChildren().iterator() |
 *	                 forall int type1 inarray assertionTypes |
 *                     (s1.getSetting( type1 ) != s2.getSetting( type1 )) implies getSetting( type1 ) == MIXED
 *
 * @invariant forall Setting s3 initerator getChildren().iterator() |
 *	              forall int type2 inarray assertionTypes |
 *	                 getSetting( type2 ) == ON implies s3.getSetting( type2 ) == ON
 *
 * @invariant forall Setting s4 initerator getChildren().iterator() |
 *	              forall int type3 inarray assertionTypes |
 *	                 getSetting( type3 ) == OFF implies s4.getSetting( type3 ) == OFF
 *
 */
public abstract class Setting implements Comparable {

//	private static int idCounter = 0;
//	private final int id;
	protected static final ArrayList errorClasses = new ArrayList();
	
	private final Setting parent;

	protected final String[] settings = { OFF, OFF, OFF };
	private final List children = new ArrayList();

	public static final int PRE       = 0;
	public static final int POST      = 1;
	public static final int INVARIANT = 2;
	protected final int[] assertionTypes = { PRE, POST, INVARIANT };

	public static final String ON    = "ON";
	public static final String OFF   = "OFF";
	public static final String MIXED = "MIXED";

	public static final FileFilter classFilter = new FileFilter() {
		public boolean accept( File file ) {
			return file.getName().endsWith( ".class" );
		}
	};

	public static final FileFilter directoryFilter = new FileFilter() {
		public boolean accept( File file ) {
			return file.isDirectory();
		}
	};
	
	public static List getErrorClasses() {
		return errorClasses;
	}
	
	public static void resetErrors() {
		errorClasses.clear();
	}
	
	private static IControlListener listener;
	private static final IControlEvent evt = new IControlEvent();
	
	protected static void fireAnalyzingSetting( Setting setting ) {
		if( listener != null ) {
			evt.setSetting( setting );
			listener.analyzingSetting( evt );
		}
	}
	
	public static void setIControlListener( IControlListener l ) {
		listener = l;
	}

/*
	public int getId() {
		return id;
	}
*/
	protected Setting( Setting parent ) {
		this.parent = parent;
//		id = idCounter++;
	}

	public Setting getParent() {
		return parent;
	}

	/**
	 * Gets the precondition setting
	 *
	 * @pre type == PRE || type == POST || type == INVARIANT
	 * @post return == ON || return == OFF || return == MIXED
	 * @return the precondition setting.
	 */
	public String getSetting( int type ) {
		String result = null;
		Iterator i = getChildren().iterator();
		if( i.hasNext() ) {
			// get the setting of the first method
			String firstChildSetting = ((Setting) i.next()).getSetting( type );
			boolean mixed = false;
			while( i.hasNext() && !mixed ) {
				if( ((Setting) i.next()).getSetting( type ) != firstChildSetting ) {
					mixed = true;
				}
			}
			if( mixed ) {
				result = MIXED;
			} else {
				result = firstChildSetting;
			}
		} else {
			result = settings[ type ];
		}
		return result;
	}

	/**
	 * Sets the precondition setting
	 *
	 * @param the precondition setting.
	 * @pre type == PRE || type == POST || type == INVARIANT
	 * @pre value == ON || value == OFF
	 * @post getSetting( type ) == value
	 */
	public void setSetting( int type, String value ) {
		Iterator i = getChildren().iterator();
		if( i.hasNext() ) {
			// get the setting of the first method
			while( i.hasNext() ) {
				((Setting) i.next()).setSetting( type, value );
			}
		} else {
			settings[ type ] = value;
		}
	}

	/**
	 * Returns settings as string. Used when control file is written
	 *
	 * @return settings as String
	 * @post return != null
	 */
	protected String getSettings() {
		throw new RuntimeException( "Not supposed to call getSettings() on" + getClass().getName() );
	}

	/**
	 * String representation with full package, class and method qualification
	 *
	 * @return a String representation
	 * @post return != null
	 */
	public abstract String getQualifiedName();

	public final String getControlFileName() {
		StringBuffer sb = new StringBuffer( getQualifiedName() ).append( getSettings() );
		return sb.toString();
	}

	/**
	 * Gets the name, that is package, class or method
	 *
	 * @return a String representation
	 * @post return != null
	 */
	public abstract String getDisplayName();

	public int compareTo( Object o ) {
		int result = 0;
		if( o instanceof Setting ) {
			Setting other = (Setting) o;
			result = getControlFileName().compareTo( other.getQualifiedName() );
		}
		return result;
	}

	public List getChildren() {
		return children;
	}

	public void addChild( Setting setting ) {
//		children.put( "" + setting.getId(), setting );
		children.add( setting );
	}
/*
	public Setting getChild( String id ) {
		return (Setting) children.get( id );
	}
*/
	public String toString() {
		return getDisplayName();
	}
}

